// The included source defines a snapshot test.
#![allow(clippy::test_attr_in_doctest)]

//! # Example: pacman-like CLI (Builder API)
//!
//! ```rust
#![doc = include_str!("../../examples/pacman.rs")]
//! ```
//!
#![doc = include_str!("../../examples/pacman.md")]
